#include <linux/module.h>
#include <linux/input.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/moduleparam.h>
#include <linux/hid.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/cdev.h>
#include "usbhid.h"
#include <asm/uaccess.h>

#define ILITEK_HID_DRIVER_NAME		"ilitek_hid"
#define ILITEK_FILE_DRIVER_NAME		"ilitek_file"
#define ILITEK_DEBUG_LEVEL		KERN_ALERT
#define ILITEK_ERROR_LEVEL		KERN_ALERT
#ifndef USB_VENDOR_ID_ILITEK
	#define USB_VENDOR_ID_ILITEK	0x222A
#endif
#ifndef USB_DEVICE_ID_ILITEK_MT
	#define USB_DEVICE_ID_ILITEK_MT	0x001E
#endif

#define ILITEK_USB_CMD_CALIBRATION      0xCC
#define ILITEK_DRIVER_VERSION "ilitek_au2v6"
#define DBG(fmt, args...)   if (DBG_FLAG)printk("%s(%d): " fmt, __func__,__LINE__,  ## args)

MODULE_AUTHOR("ilitek");
MODULE_DESCRIPTION("Android driver for ilitek touch screen");
MODULE_LICENSE("GPL");

struct touch_data{
	int x, y, z;
	int valid;
	int id;
};

struct hid_data {
	struct input_dev *input_dev;
	struct touch_data tp[10];
	int max_x;
	int max_y;
	int valid_input_register;
	int valid_hid_register;
	int support_hid_device;
	int support_touch_point;
	struct hid_device *hid_dev;
};

struct dev_data {
	dev_t devno;
	struct cdev cdev;
	struct class *class;
};

struct ilitek_data {
	struct hid_data hid;
	struct dev_data dev;
};
static struct ilitek_data ilitek;

static void ilitek_input_report(struct input_dev*, struct touch_data*, int);
static int ilitek_register_input_device(struct input_dev**, int, int, int, char*);
static int ilitek_register_hid_device(struct hid_data*);
static int ilitek_init(void);
static void ilitek_exit(void);
static int ilitek_hid_set_report(unsigned char type, unsigned char id, void *buf, int size);

static int hid_input_mapping(struct hid_device*, struct hid_input*, struct hid_field*, struct hid_usage*, unsigned long**, int*);
static int hid_input_mapped(struct hid_device*, struct hid_input*, struct hid_field*, struct hid_usage*, unsigned long**bit, int*);
static int hid_event(struct hid_device*, struct hid_field*, struct hid_usage*, __s32);
static int hid_probe(struct hid_device*, const struct hid_device_id*);
static void hid_remove(struct hid_device*);

static int ilitek_file_open(struct inode*, struct file*);
static ssize_t ilitek_file_write(struct file*, const char*, size_t, loff_t*);
static int ilitek_file_close(struct inode*, struct file*);

struct file_operations ilitek_fops =
{
	.write = ilitek_file_write,
	.open = ilitek_file_open,
	.release = ilitek_file_close,
};

static char DBG_FLAG;

static const struct hid_device_id hid_devices[] = {
	{ HID_USB_DEVICE(USB_VENDOR_ID_ILITEK, USB_DEVICE_ID_ILITEK_MT) },
	{ HID_USB_DEVICE(USB_VENDOR_ID_ILITEK, 0x0010) },
	{ HID_USB_DEVICE(USB_VENDOR_ID_ILITEK, 0x0001) },
	{ }
};
MODULE_DEVICE_TABLE(hid, hid_devices);

static const struct hid_usage_id hid_id[] = {
	{ HID_ANY_ID, HID_ANY_ID, HID_ANY_ID },
	{ HID_ANY_ID - 1, HID_ANY_ID - 1, HID_ANY_ID - 1}
};

static struct hid_driver hid_driver = {
	.name = ILITEK_HID_DRIVER_NAME,
	.id_table = hid_devices,
	.probe = hid_probe,
	.remove = hid_remove,
	.input_mapping = hid_input_mapping,
	.input_mapped = hid_input_mapped,
	.usage_table = hid_id,
	.event = hid_event,
};

static int 
ilitek_hid_set_report(
	unsigned char type,
	unsigned char id, void *buf, int size)
{
	return usb_control_msg(hid_to_usb_dev(ilitek.hid.hid_dev), 
	usb_sndctrlpipe(hid_to_usb_dev(ilitek.hid.hid_dev), 0),
	0x09, USB_TYPE_CLASS | USB_RECIP_INTERFACE, (type << 8) + id,
	0, buf, size, HZ);
}

static int 
ilitek_file_open(
	struct inode *inode, struct file *filp)
{
	printk(ILITEK_DEBUG_LEVEL "%s\n", __func__);
	return 0; 
}

static ssize_t 
ilitek_file_write(
	struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	uint8_t buffer[128]={0};
	if(count > 128){
		return -1;
	}

	memcpy(buffer, buf, count-1);
	if(strcmp(buffer, "calibrate") == 0){
		if(ilitek.hid.support_hid_device){
			if(ilitek.hid.valid_hid_register){
				if(ilitek.hid.hid_dev){
					unsigned char buf[128]={0};
					buf[0] = 0x03;
					buf[1] = ILITEK_USB_CMD_CALIBRATION;
					buf[2] = 0x01;
					if(ilitek_hid_set_report(0x02, 0x03, buf, 0x40) < 0){
						printk("%s, usb calibrate, failed\n", __func__);
					}
					else{
						printk("%s, usb calibrate, success\n", __func__);
					}
				}
			}
		}
	}else if(strcmp(buffer, "dbg") == 0){
		DBG_FLAG=!DBG_FLAG;
		printk("%s, %s message(%X).\n",__func__,DBG_FLAG?"Enabled":"Disabled",DBG_FLAG);
	}
	else{
		printk("%s, invalid command\n", __func__);
	}
	return -1;
}

static int 
ilitek_file_close(
	struct inode *inode, struct file *filp)
{
	printk(ILITEK_DEBUG_LEVEL "%s\n", __func__);
	return 0;
}

static int 
ilitek_register_input_device(
	struct input_dev **input, int support_touch_point, 
	int max_x, int max_y, char* name)
{
	struct input_dev *input_dev;
	int ret;

	*input = input_allocate_device();
	if(*input == NULL){
		printk(ILITEK_ERROR_LEVEL "%s, input device allocation error", __func__);
		return -1;
	}
	input_dev = *input;
	if(support_touch_point == 1){
		input_dev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
		input_dev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);
		input_set_abs_params(input_dev, ABS_X, 0, max_x, 0, 0);
		input_set_abs_params(input_dev, ABS_Y, 0, max_y, 0, 0);
		input_set_abs_params(input_dev, ABS_PRESSURE, 0, 255, 0, 0);
		input_set_abs_params(input_dev, ABS_TOOL_WIDTH, 0, 1, 0, 0);
		printk(ILITEK_DEBUG_LEVEL "%s, single-touch, max_x %d, max_y %d\n", __func__, max_x, max_y);
	}
	else{
		set_bit(EV_ABS, input_dev->evbit);
		input_set_abs_params(input_dev, ABS_MT_POSITION_X, 0, max_x, 0, 0);
		input_set_abs_params(input_dev, ABS_MT_POSITION_Y, 0, max_y, 0, 0);
		input_set_abs_params(input_dev, ABS_MT_TOUCH_MAJOR, 0, 255, 0, 0);
		input_set_abs_params(input_dev, ABS_MT_WIDTH_MAJOR, 0, 1, 0, 0);
		printk(ILITEK_DEBUG_LEVEL "%s, multi-touch, max_x %d, max_y %d\n", __func__, max_x, max_y);
	}
	input_dev->name = name;
	ret = input_register_device(input_dev);
	if(ret){
		printk(ILITEK_ERROR_LEVEL "%s, input register device, error\n", __func__);
		return ret;
	}
	printk(ILITEK_DEBUG_LEVEL "%s, input register device, success\n", __func__);
	return 0;
}

static int 
hid_input_mapping(
	struct hid_device *hdev, struct hid_input *hi, 
	struct hid_field *field, struct hid_usage *usage, 
	unsigned long **bit, int *max)
{
	switch(usage->hid & HID_USAGE_PAGE){
		case HID_UP_GENDESK:
			switch(usage->hid){
				case HID_GD_X:
					if(ilitek.hid.support_touch_point == 1){
						hid_map_usage(hi, usage, bit, max, EV_ABS, ABS_X);
						input_set_abs_params(hi->input, ABS_X, field->logical_minimum, field->logical_maximum, 0, 0);
					}
					else{
						hid_map_usage(hi, usage, bit, max, EV_ABS, ABS_MT_POSITION_X);
						input_set_abs_params(hi->input, ABS_MT_POSITION_X, field->logical_minimum, field->logical_maximum, 0, 0);
					}
					ilitek.hid.max_x = field->logical_maximum;
					return 1;
				case HID_GD_Y:
					if(ilitek.hid.support_touch_point == 1){
						hid_map_usage(hi, usage, bit, max, EV_ABS, ABS_Y);
						input_set_abs_params(hi->input, ABS_Y, field->logical_minimum, field->logical_maximum, 0, 0);
					}
					else{
						hid_map_usage(hi, usage, bit, max, EV_ABS, ABS_MT_POSITION_Y);
						input_set_abs_params(hi->input, ABS_MT_POSITION_Y, field->logical_minimum, field->logical_maximum, 0, 0);
					}
					ilitek.hid.max_y = field->logical_maximum;
					printk(ILITEK_DEBUG_LEVEL "%s, max_x %d, max_y %d\n", __func__, ilitek.hid.max_x, ilitek.hid.max_y);

					if(ilitek.hid.valid_input_register == 0){
						int ret = ilitek_register_input_device(&(ilitek.hid.input_dev), ilitek.hid.support_touch_point,
						ilitek.hid.max_x, ilitek.hid.max_y, ILITEK_HID_DRIVER_NAME);
						if(ret){
							printk(ILITEK_ERROR_LEVEL "%s, input register device, error\n", __func__);
						}
						else{
							ilitek.hid.valid_input_register = 1;
							printk(ILITEK_ERROR_LEVEL "%s, input register device, success\n", __func__);
						}
					}
					return 1;
			}
			return 0;
		case HID_UP_DIGITIZER:
			switch(usage->hid){
			case HID_DG_TIPSWITCH:
				hid_map_usage(hi, usage, bit, max, EV_KEY, BTN_TOUCH);
				return 1;
			case HID_DG_INRANGE:
			case HID_DG_CONFIDENCE:
			case HID_DG_CONTACTCOUNT:
			case HID_DG_CONTACTMAX:
			case HID_DG_CONTACTID:
				return -1;
			case HID_DG_TIPPRESSURE:
				hid_map_usage(hi, usage, bit, max, EV_ABS, ABS_PRESSURE);
				return 1;
			}
			return 0;
	}
	return -1;
}

static int 
hid_input_mapped(
	struct hid_device *hdev, struct hid_input *hi, 
	struct hid_field *field, struct hid_usage *usage, 
	unsigned long **bit, int *max)
{
	if(usage->type == EV_KEY || usage->type == EV_ABS){
		clear_bit(usage->code, *bit);
	}
	return 0;
}

static int 
hid_event(
	struct hid_device *hid, struct hid_field *field, 
	struct hid_usage *usage, __s32 value)
{
	static int index = 0;
	if(hid->claimed & HID_CLAIMED_INPUT){
		switch (usage->hid) {
		case HID_DG_INRANGE:
			break;
		case HID_DG_CONFIDENCE:
			//ilitek.hid.tp[index].valid = !!value;
			break;
		case HID_DG_TIPSWITCH:
			ilitek.hid.tp[index].valid = value;
			break;
		case HID_DG_TIPPRESSURE:
			ilitek.hid.tp[index].z = value;
			break;
		case HID_DG_CONTACTID:
			ilitek.hid.tp[index].id = value;
			break;
		case HID_GD_X:
			ilitek.hid.tp[index].x = value;
			//printk("x[%d]=%d",index,value);
			break;
		case HID_GD_Y:
			ilitek.hid.tp[index].y = value;
			//printk("y[%d]=%d",index,value);
			index+= 1;
			break;
		case HID_DG_CONTACTCOUNT:
			ilitek_input_report(ilitek.hid.input_dev, ilitek.hid.tp, ilitek.hid.support_touch_point);
			index = 0;
			break;
		default:
			return 0;
		}
	}

	if(hid->claimed & HID_CLAIMED_HIDDEV && hid->hiddev_hid_event){
		hid->hiddev_hid_event(hid, field, usage, value);
	}
	return 1;
}

static int 
hid_probe(
	struct hid_device *hdev, const struct hid_device_id *dev_id)
{
	int ret;
	struct hid_report *report;

	ret = hid_parse(hdev);
	if(ret){
		printk(ILITEK_ERROR_LEVEL "%s, hid parse error, ret %d\n", __func__, ret);
		return ret;
	}

	ret = hid_hw_start(hdev, HID_CONNECT_DEFAULT);
	if(ret){
		printk(ILITEK_ERROR_LEVEL "%s, hid start error, ret %d\n", __func__, ret);
		return ret;
	}

	report = hdev->report_enum[HID_FEATURE_REPORT].report_id_hash[5]; 
	if(report){
		ilitek.hid.hid_dev = hdev;
		report->field[0]->value[0] = 2;
		usbhid_submit_report(hdev, report, USB_DIR_OUT);
	}
	return ret;
}

static void 
hid_remove(
	struct hid_device *hdev)
{
	printk(ILITEK_DEBUG_LEVEL "%s, remove hid device\n", __func__);
	hid_hw_stop(hdev);
}

static void
ilitek_input_report(
    struct input_dev *input_dev,
    struct touch_data *tp, int tpnum)
{
	int i,release_count = 0;
	DBG("%s,enter,tpnum=%d\n",__func__,tpnum);
	if(tpnum == 1) {
		if(tp[0].valid) {
			input_event(input_dev, EV_ABS, ABS_X, tp[0].x);
			input_event(input_dev, EV_ABS, ABS_Y, tp[0].y);
			input_event(input_dev, EV_KEY, BTN_TOUCH, 1);
		} else {
			input_event(input_dev, EV_KEY, BTN_TOUCH, 0);
		}
	} else {
		for(i=0;i<tpnum;i++){
			if(tp[i].valid) {
				input_event(input_dev, EV_ABS, ABS_MT_TRACKING_ID, tp[i].id);
				input_event(input_dev, EV_ABS, ABS_MT_POSITION_X, tp[i].x);
				input_event(input_dev, EV_ABS, ABS_MT_POSITION_Y, tp[i].y);
				input_event(input_dev, EV_ABS, ABS_MT_TOUCH_MAJOR, 1);
				input_event(input_dev, EV_KEY, BTN_TOUCH, 1);
				input_mt_sync(input_dev);
				DBG("touch down,x%d=%d,y%d=%d\n",i,tp[i].x,i,tp[i].y);
			} else {
				input_event(input_dev, EV_ABS, ABS_MT_TRACKING_ID, tp[i].id);
				DBG("touch up ,x%d=%d,y%d=%d\n",i,tp[i].x,i,tp[i].y);
				release_count++;
				if(release_count == tpnum){
					input_event(input_dev, EV_KEY, BTN_TOUCH, 0);
					DBG("touch release\n");
				}
				input_mt_sync(input_dev);
			}
		}
	}
	input_sync(input_dev);
}

static int 
ilitek_register_hid_device(
	struct hid_data *hid)
{
	int ret;

	if(hid == NULL){
		printk(ILITEK_ERROR_LEVEL "%s, invalid parameter\n", __func__);
		return -1;
	}

	ret = hid_register_driver(&hid_driver);
	if(ret < 0){
		printk(ILITEK_ERROR_LEVEL "%s, register hid driver error, ret %d\n", __func__, ret);
		return ret;
	}
	hid->valid_hid_register = 1;
	return ret;
}

static int 
ilitek_init(
	void)
{
	int ret = 0;
    memset(&ilitek, 0, sizeof(struct ilitek_data));
    ilitek.hid.support_hid_device = 1;
	ilitek.hid.support_touch_point = 10;
	if(ilitek.hid.support_hid_device){
		ret = ilitek_register_hid_device(&ilitek.hid);
		if(ret < 0){
			printk(ILITEK_ERROR_LEVEL "%s, register hid device, error\n", __func__);
		}
	}
	
	ret = alloc_chrdev_region(&ilitek.dev.devno, 0, 1, ILITEK_FILE_DRIVER_NAME);
	if(ret){
		printk(ILITEK_ERROR_LEVEL "%s, can't alloc chrdev\n", __func__);
	}
    printk(ILITEK_DEBUG_LEVEL "%s, register chrdev(%d, %d)\n", __func__, MAJOR(ilitek.dev.devno), MINOR(ilitek.dev.devno));
	printk("Driver Version : %s\n",ILITEK_DRIVER_VERSION);
	cdev_init(&ilitek.dev.cdev, &ilitek_fops);
	ilitek.dev.cdev.owner = THIS_MODULE;
	ret = cdev_add(&ilitek.dev.cdev, ilitek.dev.devno, 1);
	if(ret < 0){
		printk(ILITEK_ERROR_LEVEL "%s, add char devive error, ret %d\n", __func__, ret);
	}
	ilitek.dev.class = class_create(THIS_MODULE, ILITEK_FILE_DRIVER_NAME);
	if(IS_ERR(ilitek.dev.class)){
		printk(ILITEK_ERROR_LEVEL "%s, creating class error\n", __func__);
	}
	device_create(ilitek.dev.class, NULL, ilitek.dev.devno, NULL, "ilitek_ctrl");
	return 0;
}

static void 
ilitek_exit(
	void)
{
	if(ilitek.hid.support_hid_device){
		if(ilitek.hid.valid_hid_register){
			printk(ILITEK_DEBUG_LEVEL "%s, unregister hid driver\n", __func__);
			hid_unregister_driver(&hid_driver);
		}
		if(ilitek.hid.valid_input_register){
			printk(ILITEK_DEBUG_LEVEL "%s, unregister hid input device\n", __func__);
			input_unregister_device(ilitek.hid.input_dev);
		}
	}
	
	cdev_del(&ilitek.dev.cdev);
	unregister_chrdev_region(ilitek.dev.devno, 1);
	device_destroy(ilitek.dev.class, ilitek.dev.devno);
    class_destroy(ilitek.dev.class);
}

module_init(ilitek_init);
module_exit(ilitek_exit);

